/************************************************************************/
/*  Routines to position input stream.					*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2000 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "$Id: position_stream.c,v 1.3 2006/03/10 21:47:07 doug Exp $ ";
#endif

#include <stdio.h>
#include <math.h>

#include "qlib2.h"

#include "params.h"
#include "st_info.h"
#include "procs.h"
#include "externals.h"

/************************************************************************/
/*  position_stream:							*/
/*	Position the input stream to the first block that does not	*/
/*	preceed the specified time.					*/
/************************************************************************/
int position_stream 
   (ST_INFO	*st_p,		/* ptr to current input stream.		*/
    INT_TIME	*stime,		/* ptr to desired start time.		*/
    INT_TIME	*etime,		/* ptr to desired end time.		*/
    int		record_num)	/* blockette 2000 record number.	*/
{
    int		nr;
    int		recnum0, recnum1, recnum2;

    while (1) {
	/* Get a new block if necessary.    */
	if (st_p->cur_hdr == NULL) {
	    /* Ensure that we have a block of data cached.		*/
	    if ( (nr = fill_buffer (st_p, st_p->blksize)) < st_p->blksize ) {
		if (nr != 0) fprintf (info, "partial block of %d bytes at end of %s\n",
				      nr, st_p->filename);
		return(EOF);
	    }
	    get_block_info (st_p);	    /* examine the block.	*/
	}
	recnum0 = (st_p->prev_hdr) ? st_p->prev_hdr->xm2 : -1;
	recnum1 = st_p->cur_hdr->xm1;
	recnum2 = st_p->cur_hdr->xm2;

	/* Selection criteria for this record.				*/
	do {
	    /* if end_time <= block start_time, return EOF.	*/
	    if ((etime != NULL) && (tdiff(*etime,st_p->cur_hdr->begtime) <= 0))
		return (EOF);

	    /* WARNING - WARNING - WARNING - WARNING - WARNING		*/
	    /* ASSUMPTION:  Each unique blockette 2000 has its own	*/
	    /* unique record number.					*/
	    if (record_num != IGNORE2000) {
		/* Search for desired blockette 2000 record number.		*/
		if ((recnum2 != IGNORE2000) && 
		    (GT2000(recnum2, record_num) || 
		     (recnum0 >= 0 && fabs(recnum0-recnum1) > block_tol)))
		    return(0);
		break;	/* next record, please */
	    }

	    /* If no time is specified, return the current block.	*/
	    if (stime == NULL) return(0);

	    /* If start_time <= block endtime, return the current block.*/
	    if (tdiff(*stime,st_p->cur_hdr->endtime) <= 0 ) return (0);

	} while (0);


	/* Otherwise, get the next block.   */
	if (st_p->prev != NULL) free_data_hdr (st_p->prev_hdr);
	st_p->prev_hdr = st_p->cur_hdr;
	st_p->cur_hdr = NULL;
/*::	st_p->iob->avail = st_p->iob->free;*/
	st_p->iob->avail += st_p->blksize;
	st_p->used = 1;
    }
    return(0);
}

