/************************************************************************/
/*  Routines to process current input record.				*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2000 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "$Id: process_block.c,v 1.4 2000/11/04 14:39:10 doug Exp $ ";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "qlib2.h"

#include "params.h"
#include "st_info.h"
#include "procs.h"
#include "externals.h"

static DATA_HDR *last_hdr;	/* header of last packet used.		*/

/************************************************************************/
/*  process_block:							*/
/*	Process input record for output.				*/
/*									*/
/*  This may involve any or all of the following:			*/
/*  1.  Detecting and discarding duplicate records.			*/
/*  2.  Trimming data from either end of the record in order to satisfy	*/
/*	time interval or in order to splice data from different files	*/
/*	into the output stream (requires unpacking/repacking).		*/
/*  3.	Re-blocking the data into a different output blksize.  If the	*/
/*	output blksize is smaller than the input blksize, it requires	*/
/*	unpacking the data and repacking it into multiple records.	*/
/*  4.  Determining what blockettes get put into the output stream.	*/
/*  5.	Appending the resulting data and blockettes to output stream.	*/
/************************************************************************/
void process_block 
   (ST_INFO	*st_p,		/* ptr to input stream info.		*/
    int		new_stream)	/* boolean flag - true if new stream.	*/
{
    INT_TIME	    exptime, *p_last_time;
    ST_INFO	    *out = st_head_out.next;
    int		    trim;
    int		    ok_to_merge;
    char	    debug_buf[65536];
    static	    same_stream_as_last = 0;
    int		    discard_packet = 0;
    int		    orig_blksize;
    int		    h;
    RECORD_LIST	    record_list[100];
    BS		    *bs;		/* ptr to blockette structure.	*/
    BLOCKETTE_HDR   *bh;		/* ptr to blockette hdr.	*/
    IOB		    *prev_iob;		/* ptr to previous iob.		*/
    int		    trim_through;	/* b2000 recnum to trim through	*/

    if (new_stream) same_stream_as_last = 0;
    memset(record_list, 0, sizeof(record_list));

    /* Save original block info in record_list[0], and duplicate it in	*/
    /* record_list[1].  We need to do this because we may be forced to	*/  
    /* unpack and repack the data from this block into multiple blocks	*/
    /* due to trimming the data, changing blocksize, etc.		*/
    /* Process block starting with record_list[1].			*/
    /* We keep record_list[0] unmodified as the original input.		*/
    orig_blksize = st_p->blksize;
    record_list[0].hdr = NULL;
    record_list[0].iob = NULL;
    record_list[1].hdr = st_p->cur_hdr;
    record_list[1].iob = st_p->iob;    
    h = 1;
    debug_buf[0] = '\0';

    while ((st_p->cur_hdr = record_list[h].hdr, st_p->iob = record_list[h].iob) != NULL) {
	/* Ensure that all headers for the output stream are 		*/
	/* allocated and initialized.					*/
	out = st_head_out.next;
	if (out->sum_hdr == NULL) make_out_sum_hdr(out, st_p);
	if (out->cur_hdr == NULL) make_out_cur_hdr(out);
	if (debug(DEBUG_HDRS)) {
	    dump_hdr (record_list[h].hdr, debug_buf+strlen(debug_buf), date_fmt);
	    dump_blk (record_list[h].hdr, debug_buf+strlen(debug_buf));
	    fprintf (info, "strlen(debug_buf) = %d\n", strlen(debug_buf));
	    fflush (info);
	}

	/*	Perform sanity checks on the input stream.		*/
	if (! same_type_data_stream (record_list[h].hdr, out->sum_hdr)) {
	    fprintf (info, "Error: input block for different stream than output stream\n");
	    fprintf (info, "output = %s.%s.%s.%s, ",
		    out->sum_hdr->station_id, out->sum_hdr->network_id,
		    out->sum_hdr->channel_id, out->sum_hdr->location_id);
	    fprintf (info, "input = %s.%s.%s.%s\n",
		    record_list[h].hdr->station_id, record_list[h].hdr->network_id,
		    record_list[h].hdr->channel_id, record_list[h].hdr->location_id);
	    if (debug(DEBUG_HDRS)) fprintf (info, debug_buf);
	    exit(1);
	}

	if (record_list[h].hdr->num_samples > 0) {
	    fprintf (info, "Error: Cannot process records with blockette 2000 and data samples\n");
	    fprintf (info, "input = %s.%s.%s.%s, num_samples = %d\n",
		    record_list[h].hdr->station_id, record_list[h].hdr->network_id,
		    record_list[h].hdr->channel_id, record_list[h].hdr->location_id,
		    record_list[h].hdr->num_samples);
	    if (debug(DEBUG_HDRS)) fprintf (info, debug_buf);
	    exit(1);
	}

	/* Determine whether we have to discard any blockettes. */
	trim = 0;
	trim_through = IGNORE2000;

	/* 1.  If this is the same input stream as last record,		*/
	/* and if start time of record <= start time of last record,	*/
	/* and last b2000 recnum <= output b2000 recnum,		*/
	/* then record is a duplicate (or out of order) so discard it.	*/

	if (same_stream_as_last && st_p->prev_hdr != NULL &&
	    tdiff(st_p->cur_hdr->begtime,st_p->prev_hdr->begtime)<=0. &&
	    LE2000(st_p->cur_hdr->xm2,out->cur_hdr->xm2)) {
	    ++discard_packet;
	}

	else if (LE2000(st_p->cur_hdr->xm1,out->cur_hdr->xm2) &&
		 GE2000(st_p->cur_hdr->xm2,out->cur_hdr->xm2)) {
	    /* Discard blockettes up to and including the last output	*/
	    /* blockette number.					*/
	    trim = 1;
	    trim_through = out->cur_hdr->xm2;
	}
	else if (st_p->prev_hdr != NULL && 
		 tdiff(st_p->cur_hdr->begtime,st_p->prev_hdr->begtime)>= 0. &&
		 st_p->cur_hdr->xm1 == 0) {
	    /* Assume restart of blockette generating program.		*/
	    /* Use all blockettes in current record.			*/
	    trim = 0;
	}
	if (st_p->cur_hdr->blksize > out->blksize) {
	    trim = 1;
	}

	if (trim && (! discard_packet)) {
	    /* Save original record and IOB if not already saved.	*/
	    if (record_list[0].hdr == NULL) {
		record_list[0].hdr = dup_data_hdr(record_list[h].hdr);
		record_list[0].iob = dup_iob(record_list[h].iob);
	    }
	    trim_data (st_p, trim_through, out->blksize, &record_list[h]);
	    continue;
	}
	else {
	    if (out->blksize < st_p->cur_hdr->blksize) {
		/* Use trim_data to reblock the input data */
		trim_data (st_p, IGNORE2000, out->blksize, &record_list[h]);
		continue;
	    }
	}

	if (! discard_packet) {
	    if (file_duration) {
		if (! init_duration) compute_duration (record_list[h].hdr);
	    }
	    /* Determine whether this is consecutive data.		*/
	    if (out->cur_hdr->xm2 != IGNORE2000 &&
		(NEXT2000(out->cur_hdr->xm2) != st_p->cur_hdr->xm1 
/*::		 && st_p->cur_hdr->xm1 != 1 ::*/
		 )) {
		/* Non-consecutive data.  Flush output.			*/
		flush_output(out);
		print_stream_summary(out);
		out->min_slew = out->max_slew = out->max_step = out->total_slew = 0;
		out->min_slew = out->max_slew = out->max_step = out->total_slew = 0;
		/* Deallocate header structures and reallocate them for	*/
		/* a new contiguous segment.				*/
		free_data_hdr (out->sum_hdr);
		free_data_hdr (out->cur_hdr);
		free_data_hdr (out->prev_hdr);
		out->sum_hdr = out->cur_hdr = out->prev_hdr = NULL;
		continue;
	    }

	    add_required_blockettes(record_list[h].hdr);
	    delete_unwanted_blockettes(record_list[h].hdr);
	    ok_to_merge =  merge_check (out->cur_hdr, record_list[h].hdr);
	    if (!ok_to_merge) flush_output(out);
	    append_to_output (st_p, 0);
	    st_p->used = 1;
	}
	else {
	    /*  We are discarding the entire packet.			*/
	    if (debug(DEBUG_TIME)||debug(DEBUG_STREAM)||debug(DEBUG_BLOCK)) 
		fprintf (info, "discarding entire input packet\n");
	    ++discard_packet;
	    break;
	}
	/* Ensure we maintain a copy of the initial record.		*/
	if (record_list[0].hdr == NULL) {
	    record_list[0].hdr = record_list[h].hdr;
	    record_list[0].iob = record_list[h].iob;
	    record_list[h].hdr = st_p->cur_hdr = NULL;
	    record_list[h].iob = st_p->iob = NULL;
	}
	/* Free current record and iob (if not initial record).		*/
	if (record_list[h].hdr != NULL) {
	    free_data_hdr (record_list[h].hdr);
	    free_iob(record_list[h].iob);
	    record_list[h].hdr = st_p->cur_hdr = NULL;
	    record_list[h].iob = st_p->iob = NULL;
	}
	h++;
    }

    /* Final cleanup.							*/
    if (debug(DEBUG_HDRS)) fprintf (info, debug_buf);

    /* Ensure we maintain a copy of the initial record.		*/
    if (record_list[0].hdr == NULL) {
	record_list[0].hdr = record_list[h].hdr;
	record_list[0].iob = record_list[h].iob;
	record_list[h].hdr = st_p->cur_hdr = NULL;
	record_list[h].iob = st_p->iob = NULL;
	h++;
    }
    /* Free any remaining headers in header list.			*/
    while (record_list[h].hdr != NULL) {
	free_data_hdr(record_list[h].hdr);
	free_iob(record_list[h].iob);
	h++;
    }

    /* Update input header structures.					*/
    /* Ensure that all created headers are freed.			*/
    st_p->cur_hdr = record_list[0].hdr;
    st_p->iob = record_list[0].iob;
    st_p->blksize = orig_blksize;
    st_p->iob->avail = st_p->blksize;
    if (! discard_packet) {
	if (st_p->prev_hdr != NULL) free_data_hdr (st_p->prev_hdr);
	st_p->prev_hdr = st_p->cur_hdr;
    }
    st_p->cur_hdr = NULL;
    st_p->used = 1;
	
    /* Free databuff and diffbuff, since their size is block dependent.	*/
    free (databuff);
    databuff = NULL;
    free (diffbuff);
    diffbuff = NULL;
    return;
}
