/************************************************************************/
/*  Utility routines to process current input record.			*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2000 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "$Id: process_utils.c,v 1.4 2009/05/14 23:33:55 doug Exp $ ";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <math.h>

#include "qlib2.h"

#include "params.h"
#include "st_info.h"
#include "procs.h"
#include "externals.h"

#define	ANY_EVENT_FLAGS (ACTIVITY_BEGINNING_OF_EVENT | \
			     ACTIVITY_END_OF_EVENT | \
			     ACTIVITY_EVENT_IN_PROGRESS)

/************************************************************************/
/*  write_seed_hdr:							*/
/*	Output seed data block.						*/
/************************************************************************/
void write_seed_hdr
   (ST_INFO	*st_p,		/* ptr to input stream struct.		*/
    SDR_HDR	*sh)		/* ptr to SEED Data Header.		*/
{
    DATA_HDR	*p = st_p->cur_hdr;
    BS	*bs;
    BLOCKETTE_1000 *b;
    int swapflag;
    int stmp;
    int itmp;

    /* Set up SEED header and data blockettes.				*/
    /*  Set station_id, channel_id, network_id based on version number	*/
    /*  and cmdline options.						*/
    if (station_id) strcpy (p->station_id, station_id);
    if (channel_id) strcpy (p->channel_id, channel_id);
    if (location_id) strcpy (p->location_id, location_id);
    if (seed_version >= 2.3) {
	if (network_id) {
	    strcpy (p->network_id, network_id);
	}
	else if (p->network_id[0]==0 || 
		 (p->network_id[0] == ' ' && p->network_id[1] == ' ')) {
	    strcpy (p->network_id, default_network_id);
	}
    }
    else {
	p->network_id[0] = p->network_id[1] = p->network_id[2] = 0;
    }

    init_sdr_hdr (sh, p, NULL);
    update_sdr_hdr (sh, p);
}

/************************************************************************/
/*  merge_flags:							*/
/*	Boolean function that determines whether flags from 2		*/
/*	different blocks can be safely merged.				*/
/************************************************************************/
int merge_flags 
   (DATA_HDR	*oh,		/* ptr to output record DATA_HDR.	*/
    DATA_HDR	*ih)		/* ptr to intput record DATA_HDR.	*/
{
    SEED_UWORD o_activity_flags = (SEED_UWORD) oh->activity_flags & 
	~ ACTIVITY_BEGINNING_OF_EVENT;

    /*	An ACTIVITY_BEGINNING_OF_EVENT flag on input header should	*/
    /*	a new block, but it should be ignored in the output header	*/
    /*	flag when compared to the input header flag.			*/

    if ((o_activity_flags != ih->activity_flags) ||
	(oh->io_flags != ih->io_flags) ||
	(oh->data_quality_flags != ih->data_quality_flags)) return (0);
    return(1);	    /* OK to merge. */
}

/************************************************************************/
/*  merge_check -							*/
/*	Determine whether it is OK to merge data & blockettes from the 	*/
/*	current block with data & blockettes already queued for output.	*/
/*	Assume that any required blockettes already exist.		*/
/*	Return 1 for OK, 0 for not OK.					*/
/************************************************************************/
int merge_check
   (DATA_HDR	*oh,		/* ptr to output record DATA_HDR.	*/
    DATA_HDR	*ih)		/* ptr to input record DATA_HDR.	*/
{
    int status;
    BS *bs, *obs;
    int type;
    int bl_len = 0;
    int b1000_found = 0;
    int found;
    int tmp_first_data;
    MS_ATTR oh_ms_attr, ih_ms_attr;
    
    /* If we are splitting output based on duration intervals:		*/
    /* Assume that any splitting of records to prevent a record from	*/
    /* spanning a duration (if exact_trim is set) has been done.	*/
    /* Check to see if this the start of this record falls within the	*/
    /* current duration.  If not, return false.				*/
    if (file_duration && after_duration (ih)) return (0);

    /* Compute current number of bytes for blockettes in output header.	*/
    for (bs=oh->pblockettes; bs != NULL; bs=bs->next) {
	type = bs->type;
	if (type == 1000) b1000_found = 1;
	bl_len += bs->len + ((bs->len%4) ? 4-(bs->len%4) : 0);
    }
    /* Check to see how many bytes of new blockettes this input header	*/
    /* would add to the output header.					*/
    /* An input blockette is considered new if:				*/
    /*	1.  A output blockette of the same number is not found.		*/
    /*	2.  An output blockette of the same number is found, but is not */
    /*	    a per_record_blockette and the contents are different than	*/
    /*	    input blockette.						*/
    for (bs = ih->pblockettes; bs != NULL; bs = bs->next) {
	type = bs->type;
	if (find_pblockette(oh,oh->pblockettes,type) == NULL) {
	    /* new blockette for output.	*/
	    bl_len += bs->len + ((bs->len%4) ? 4-(bs->len%4) : 0);
	}
	else if (is_per_record_blockette(type)) { 
	    continue;
	}
	else {
	    /* Search ALL of the existing output blockettes to determine*/
	    /* whether this is a duplicate blockette or not.  */
	    for (found=0, obs=oh->pblockettes;
		 (obs=find_pblockette(oh,obs,type)) && !(found=!blockettecmp(bs,obs));
		 obs=obs->next) ;
	    if (! found) bl_len += bs->len + ((bs->len%4) ? 4-(bs->len%4) : 0);
	}
    }
    /* Temporarily set first_data in oh to account for space for the	*/
    /* current and proposed blockettes, which is required to compute	*/
    /* the space attributes for the output block being constructed.	*/
    tmp_first_data = oh->first_data;
    oh->first_data = 48 + bl_len;
    oh_ms_attr = get_ms_attr (oh);
    oh->first_data = tmp_first_data;
    
    ih_ms_attr = get_ms_attr (ih);

    status = (((oh->data_type == UNKNOWN_DATATYPE && oh_ms_attr.nbytes == 0) || 
	       (oh->data_type == ih->data_type)) &&
	      (ih_ms_attr.nbytes <= oh_ms_attr.bytelimit-oh_ms_attr.nbytes) && 
	      ((oh_ms_attr.nbytes == 0) || (merge_flags (oh, ih))));
    return (status);
}
    
/************************************************************************/
/*  append_to_output:							*/
/*	Append input to output.						*/
/************************************************************************/
void append_to_output 
   (ST_INFO	*st_p,		/* ptr to ouput stream struct.		*/
    int		slew)		/* slew between this and prev record.	*/
{
    ST_INFO	*out = st_head_out.next;
    DATA_HDR	*oh = out->cur_hdr;
    DATA_HDR	*ih = st_p->cur_hdr;
    DATA_HDR	*sh = out->sum_hdr;
    int		i_offset = ih->first_data;
    int		i_len;
    int		o_offset;
    int		seconds, usecs, usecs_per_point;
    FRAME	*p_fr1, *p_fr2;
    char	*str;
    int i;
    BS		*bs;
    int		num_samples;	/* number of blockette 2000s in record.	*/
    MS_ATTR	ih_ms_attr, oh_ms_attr;
    char	*filename;

    /* NOTE:  in oh and sh, num_samples =  number of b2000 blockettes.	*/
    /* NOTE:  we have to fill in first_data to point to space AFTER	*/
    /* blockettes - otherwise get_ms_attr() will return the wrong info.	*/

    /* Ensure that the data_type is set for output.			*/
    if (oh->data_type == UNKNOWN_DATATYPE) oh->data_type = ih->data_type;

    /* Ensure that input and output data_types are the same, and that	*/
    /* the byte order are the same.					*/
    if (ih->data_type != oh->data_type) {
	fprintf (stderr, "Error: input data_type = %d, output data_type = %d\n",
		 ih->data_type, oh->data_type);
	exit(1);
    }

    /* This check is not required as long as we restrict the channel	*/
    /* to ONLY b2000 and no data samples.  Check made in process_block.	*/
    /*::
    if (ih->data_wordorder != oh->data_wordorder) {
	fprintf (stderr, "Error: input data_wordorder = %d, output data_wordorder = %d\n",
		 ih->data_wordorder, oh->data_wordorder);
	exit(1);
    }
    ::*/

    /* Save input data hdr as the last data hdr seen, even if we do not	*/
    /* use the output data.  last_hdr is used by the find_best_block	*/
    /* function to determine which file to use for input.		*/
    if (out->last_hdr) free_data_hdr (out->last_hdr);
    out->last_hdr= dup_data_hdr(st_p->cur_hdr);
    if (event_only_flag && ! (st_p->cur_hdr->activity_flags & ANY_EVENT_FLAGS)) {
	return;
    }

    /* If this record starts after the current duration, we need to	*/
    /* close the existing file (if open), and open a new file.		*/
    if (file_duration && after_duration (ih)) {
	if (out) {
	    /* We should not have any pending output.  Check to be sure.*/
	    if (oh->num_samples > 0 || oh->num_blockettes > 0) {
		fprintf (info, "Error: unflushed output found\n");
		exit (1);
	    }
	    print_stream_summary(out);
	    close_stream (out);
	    out = NULL;
	}
    }

    /* If file is not open, we need to compute a new duration and open	*/
    /* open a new file based on the duration.				*/
    if (out == NULL || out->iob->deferred) {
	compute_duration (ih);
	filename = build_filename (st_p->cur_hdr, filename_pattern);
	if (out) close_stream (out);
	out = open_stream (&st_head_out, filename, outmode);
	if (out->sum_hdr == NULL) make_out_sum_hdr(out, st_p);
	if (out->cur_hdr == NULL) make_out_cur_hdr(out);
    }
    oh = out->cur_hdr;
    sh = out->sum_hdr;
	
    /* Count the number of blockette 2000s in the input record.		*/
    num_samples = 0;
    for (bs = ih->pblockettes; bs != NULL; bs = bs->next) {
	if (bs->type == 2000) ++num_samples;
    }

    /* Copy any new blockettes from the input to the output hdr.	*/
    merge_blockettes (oh, ih);
    if (oh->xm1 == IGNORE2000) oh->xm1 = ih->xm1;
    if (sh->xm1 == IGNORE2000) sh->xm1 = ih->xm1;
    oh->xm2 = ih->xm2;
    sh->xm2 = ih->xm2;

    /* Get attributes of input and output data block.			*/
    ih_ms_attr = get_ms_attr (ih);
    oh_ms_attr = get_ms_attr (oh);
    i_len = ih_ms_attr.nbytes;

    /* No data samples to copy.						*/

    /* Update the flags and other header info (including flags).	*/
    /* Ensure that ACTIVITY_TIME_GAP is set ONLY from first data block	*/
    /* put in output block.						*/
    /* Calculate endtime, since we may have chosen to ignore small	*/
    /* time jitters.							*/
    oh->io_flags |= ih->io_flags;
    oh->data_quality_flags |= ih->data_quality_flags;
    oh->num_data_frames += ih->num_data_frames;
    oh->data_type = ih->data_type;
    
    if (oh->num_samples == 0) {
	oh->activity_flags |= ih->activity_flags;
	oh->hdrtime = ih->hdrtime;
	oh->begtime = ih->begtime;
	oh->num_ticks_correction = ih->num_ticks_correction;
	if (IS_STEIM_COMP(ih->data_type)) oh->x0 = ih->x0;
    }
    else {
	oh->activity_flags |= (ih->activity_flags & (~ ACTIVITY_TIME_GAP));
    }

    oh->num_samples += num_samples;
    oh->endtime = oh->begtime;

    /* Final blockette update.  Update any blockettes that have 	*/
    /* block-specific info.						*/
    if ((bs=find_blockette(oh, 1000))) {
	/* Ensure we have proper output blocksize in the blockette.	*/
	BLOCKETTE_1000 *b = (BLOCKETTE_1000 *) bs->pb;
	b->data_rec_len = roundoff(log2((double)output_blksize));
    }
    if ((bs=find_blockette(oh, 1001))) {
	/* Ensure we have proper # of frames in the blockette.		*/
	BLOCKETTE_1001 *b = (BLOCKETTE_1001 *) bs->pb;
	b->frame_count = 0;
    }

    /*	Update the information in the output's sum_hdr.			*/
    /*	This header provides a running summary of information for data	*/
    /*	that has been directed to the output stream.  This INCLUDES	*/
    /*	data that is placed in the output buffer but not yet written.	*/
    /*	It is therefore updated NOW, NOT when the buffer is written.	*/
    /*	This provides a uniform place for begtime, endtime, num_samples,*/
    /*	and other info that is needed to reflect the state of data	*/
    /*	that has been directed for output.				*/
    if (sh->num_samples == 0) {
	sh->begtime = ih->begtime;
    }
    sh->endtime = oh->endtime;
    sh->num_samples += num_samples;

    /*	Keep track of time slew information.			*/
    slew = 0;
    out->total_slew += slew;
    out->max_slew = MAX(out->total_slew, out->max_slew);
    out->min_slew = MIN(out->total_slew, out->min_slew);
    out->max_step = MAX(out->max_step, abs(slew));
}

/************************************************************************/
/*  pad_out:								*/
/*	Pad the output buffer (if not empty).				*/
/************************************************************************/
int pad_out
   (ST_INFO	    *out,	/* ptr to output stream struct.		*/
    int		    data_bytes,	/* # of data bytes in output record.	*/
    int		    pad_len)	/* # of bytes to pad record.		*/
{
    IOB		    *iob = out->iob;
    DATA_HDR	    *ch = out->cur_hdr;

    if (ch == NULL || (ch->pblockettes == NULL && ch->num_samples == 0))
	return(0);
    memset ((void *)(iob->buf + ch->first_data +  data_bytes),
	    0, pad_len);
    return (0);
}

/************************************************************************/
/*  flush_output:							*/
/*	Flush any buffered output.					*/
/*	Since our total header + blockettes may now exceed the space	*/
/*	originally allocated for in the io buffer, we will need to	*/
/*	temporarily move the data in the i/o buffer to a temp buffer,	*/
/*	construct the output SEED header in the i/o buffer, and then	*/
/*	copy the data back into the i/o buffer for output.		*/
/************************************************************************/
void flush_output
   (ST_INFO	    *out)	/* ptr to output stream struct.		*/
{
    IOB		    *iob = out->iob;
    DATA_HDR	    *ch = out->cur_hdr;
    DATA_HDR	    *sh = out->sum_hdr;
    char	    buf[MAX_BLKSIZE];
    BS		    *bs;
    int		    nbytes_of_data;
    int		    bl_len = 0;
    int		    pad_len;
    int		    align;
    int		    save_num_samples;
    MS_ATTR	    ms_attr;

    if (ch == NULL) return;
    /* Compute current number of bytes for blockettes in output header.	*/
    for (bs=ch->pblockettes; bs != NULL; bs=bs->next) {
	bl_len += bs->len + ((bs->len%4) ? 4-(bs->len%4) : 0);
    }
    /* NOTE: explicitly set first_data for padding operation. */
    /* NOTE: save num_samples (= number of opaque blockettes). */
    ch->first_data = 48 + bl_len;
    save_num_samples = ch->num_samples;
    ch->num_samples = 0;
    ms_attr = get_ms_attr (ch);
    pad_len = out->blksize - ch->first_data;
    pad_out(out, ms_attr.nbytes, pad_len);
    nbytes_of_data = 0;		/* No data! */

    if (ch != NULL && bl_len > 0) {
	memcpy (buf, out->iob->buf+ch->first_data, nbytes_of_data);
	memset (out->iob->buf, 0, out->blksize);    /* clear output buf.*/
	ch->first_data = 48 + bl_len;
	if (align > 0) {
	    ch->first_data = (int)((ch->first_data+align-1)/align) * align;
	}
	write_seed_hdr(out, (SDR_HDR *)out->iob->buf);
	memcpy (out->iob->buf+ch->first_data, buf, nbytes_of_data);
	if (debug(DEBUG_ASCII)) write_ascii_stream (out);
	else write_stream (out);
	ch->num_samples = save_num_samples;
	if debug(DEBUG_OUT) {
	    fprintf (info, "out: %s.%s.%s.%s, block %d, %d blockettes (%s)\n",
		     ch->station_id, ch->network_id, ch->channel_id, ch->location_id, 
		     ch->seq_no, ch->num_samples,
		     time_to_str(ch->begtime,date_fmt));
	}

	/* Release old prev_hdr (if any), move cur_hdr to prev_hdr,	*/
	/* and get a new empty current header.				*/
	/* Set the data_type in the new header to current data_type.	*/
	if (out->prev_hdr != NULL) free_data_hdr(out->prev_hdr);
	out->prev_hdr = out->cur_hdr;
	out->cur_hdr = NULL;
	make_out_cur_hdr(out);
	out->cur_hdr->data_type = out->prev_hdr->data_type;
    }
}

/************************************************************************/
/*  hdr_cmp:								*/
/*	Compare 2 DATA_HDR structures (except for pblockette field).	*/
/*	Return 0 if identical, non-zero otherwise.			*/
/************************************************************************/
int hdr_cmp
   (DATA_HDR	*hdr1,		/* ptr to first DATA_HDR.		*/
    DATA_HDR	*hdr2)		/* ptr to second DATA_HDR.		*/
{
    DATA_HDR h1 = *hdr1;
    DATA_HDR h2 = *hdr2;
    h1.pblockettes = h2.pblockettes = (BS *)NULL;
    return (memcmp((void *)&h1,(void *)&h2,sizeof(DATA_HDR)));
}

/************************************************************************/
/*  same_type_data_stream:						*/
/*	Boolean function to compare stream characteristics.		*/
/************************************************************************/
int same_type_data_stream 
   (DATA_HDR	*ih,		/* ptr to first DATA_HDR.		*/
    DATA_HDR	*oh)		/* ptr to second DATA_HDR.		*/
{
    if ((strcmp(ih->station_id, oh->station_id) != 0) ||
	(strcmp(ih->channel_id, oh->channel_id) != 0) ||
	!(((ih->sample_rate == oh->sample_rate) &&
	   (ih->sample_rate_mult == oh->sample_rate_mult)) ||
	  ((ih->sample_rate == 0 && ih->num_samples == 0) || 
	   (oh->sample_rate == 0 && oh->num_samples == 0)))) return (0);
    else return (1);
}

/************************************************************************/
/*  make_out_sum_hdr:							*/
/*	Alloc and init summary output header struct.			*/
/************************************************************************/
void make_out_sum_hdr
   (ST_INFO	*out,		/* output structure.			*/
    ST_INFO	*st_p)		/* input structure.			*/
{
    /*	Make and fill in summary header if we don't have one yet.	*/
    if (out->sum_hdr == NULL) {
	out->sum_hdr = new_data_hdr();
	/*  Copy info from current input stream to summary header.	*/
	strcpy (out->sum_hdr->station_id, st_p->cur_hdr->station_id);
	strcpy (out->sum_hdr->location_id, st_p->cur_hdr->location_id);
	strcpy (out->sum_hdr->channel_id, st_p->cur_hdr->channel_id);
	strcpy (out->sum_hdr->network_id, st_p->cur_hdr->network_id);
	out->sum_hdr->sample_rate = st_p->cur_hdr->sample_rate;
	out->sum_hdr->sample_rate_mult = st_p->cur_hdr->sample_rate_mult;
	out->sum_hdr->blksize = output_blksize;
	/*  Fill in stream header structure also from summary header.	*/
	strcpy (out->station, out->sum_hdr->station_id);
	strcpy (out->channel, out->sum_hdr->channel_id);
	out->rate = out->sum_hdr->sample_rate;
	out->rate_mult = out->sum_hdr->sample_rate_mult;
	out->hdr_type = SDR_HDR_TYPE;
	out->blksize = output_blksize;
	out->sum_hdr->xm1 = IGNORE2000;
	out->sum_hdr->xm2 = IGNORE2000;

    }
}

/************************************************************************/
/*  make_out_cur_hdr:							*/
/*	Alloc and init current output header struct.			*/
/************************************************************************/
void make_out_cur_hdr
   (ST_INFO	*out)		/* ptr to output stream struct.		*/
{
    /*	Make and fill in current header if we don't have one yet.	*/
    if (out->cur_hdr == NULL) {
	out->cur_hdr = new_data_hdr();
	strcpy (out->cur_hdr->station_id, out->sum_hdr->station_id);
	strcpy (out->cur_hdr->location_id, out->sum_hdr->location_id);
	strcpy (out->cur_hdr->channel_id, out->sum_hdr->channel_id);
	strcpy (out->cur_hdr->network_id, out->sum_hdr->network_id);
	out->cur_hdr->sample_rate = out->sum_hdr->sample_rate;
	out->cur_hdr->sample_rate_mult = out->sum_hdr->sample_rate_mult;
	out->cur_hdr->first_data = 0;	    /* NOTE: was 48 */
	out->cur_hdr->blksize = output_blksize;
	out->cur_hdr->seq_no = (out->prev_hdr == NULL) ? 1 : 
	    out->prev_hdr->seq_no + 1;
	out->cur_hdr->xm1 = IGNORE2000;
	out->cur_hdr->xm2 = IGNORE2000;
    }
}

/************************************************************************/
/*  is_per_record_blockette:						*/
/*	Check if specified blockette is a per_record blockette.		*/
/*  return:								*/
/*	1 if blockette should per_record blockette, 0 otherwise.	*/
/************************************************************************/
int is_per_record_blockette 
   (int		n)		/* blockette number.			*/
{
    int i;
    for (i=0; per_record_blockette[i].blockette != 0; i++) {
	if (n == per_record_blockette[i].blockette) return (1);
    }
    return (0);
}

/************************************************************************/
/*  strip_blockette:							*/
/*	Check if specified blockette should be stripped from record.	*/
/*  return:								*/
/*	1 if blockette should be stripped, 0 otherwise.			*/
/************************************************************************/
int strip_blockette 
   (int		n)		/* blockette number.			*/
{
    int i;
    if (stripped_blockette == NULL) return (0);
    for (i=0; stripped_blockette[i].first != 0; i++) {
	if (stripped_blockette[i].first == -1 || 
	    (n >= stripped_blockette[i].first && n <= stripped_blockette[i].last))
	    return (1);
    }
    return (0);
}

/************************************************************************/
/*  add_required_blockettes:						*/
/*	Add any required blockettes to this hdr.			*/
/*  return:								*/
/*	0 on success.							*/
/************************************************************************/
int add_required_blockettes 
   (DATA_HDR	*hdr)		/* DATA_HDR for record.			*/
{
    /* Currently only blockette 1000 is required for seed_version>=2.3	*/
    if (seed_version >= 2.3 && ! strip_blockette(1000) && 
	find_blockette(hdr, 1000) == NULL) {
	BLOCKETTE_1000 b1000;
	b1000.hdr.type = 1000;
	b1000.hdr.next = 0;
	b1000.format = hdr->data_type;
	b1000.word_order = SEED_BIG_ENDIAN;
	b1000.data_rec_len = roundoff(log2((double)hdr->blksize));
	b1000.reserved = 0;
	add_blockette (hdr, (char *)&b1000, 1000, sizeof(BLOCKETTE_1000), 
		       hdr->hdr_wordorder, 0);
    }
    return (0);
}

/************************************************************************/
/*  delete_unwanted_blockettes:						*/
/*	Delete any blockettes from hdr that should not be in output.	*/
/*  return:								*/
/*	0 on success.							*/
/************************************************************************/
int delete_unwanted_blockettes 
   (DATA_HDR	*hdr)		/* DATA_HDR for record.			*/
{
    BS *bs, *delete_bs;
    int bl_type;
    for (bs=hdr->pblockettes,delete_bs=NULL;
	 bs!=NULL; 
	 bs=bs->next,delete_pblockette(hdr,delete_bs),delete_bs=NULL) {
	bl_type = bs->type;
	if (strip_blockette(bl_type)) delete_bs = bs;
    }
    return (0);
}

/************************************************************************/
/*  merge_blockettes:							*/
/*	Merge blockettes from input hdr to output hdr, making sure that:*/
/*	1.  We have only a single copies of per_record_blockettes.	*/
/*	2.  We have only a single unique copy of each unique blockette.	*/
/*  return:								*/
/*	0 on success.							*/
/************************************************************************/
int merge_blockettes 
   (DATA_HDR	*oh,		/* output record DATA_HDR.		*/
    DATA_HDR	*ih)		/* input record DATA_HDR.		*/
{
    BS *bs, *bs2;
    int bl_type, bl_pos;
    /* Add any new input blockettes to the output blockette list.	*/
    /* Allow unique copies of blockettes that are not specific		*/
    /* to a data record.						*/
    for (bs = ih->pblockettes; bs != NULL; bs=bs->next) {
	bl_type = bs->type;
	bl_pos = (bl_type == 1000) ? 0 : -1;
	if (strip_blockette(bl_type)) continue;
	if ((bs2=find_blockette(oh,bl_type)) == NULL ||
	    ((! is_per_record_blockette(bl_type)) &&
	     blockettecmp(bs,bs2) != 0)) {
	    add_blockette (oh, bs->pb, bs->type, bs->len, bs->wordorder, bl_pos);
	}
    }
    return (0);
}

#define	MAXODHSTRLEN	8192
/************************************************************************/
/*  write_ascii_stream:							*/
/*	Output data from block in ascii.				*/
/************************************************************************/
void write_ascii_stream
   (ST_INFO	*out)		/* ptr to output stream struct.		*/
{
    BS *bs;
    BLOCKETTE_HDR *bh;
    char *p, str[MAXODHSTRLEN+1];
    int i;
    int bl_type, bl_len;

    if (my_wordorder < 0) get_my_wordorder();
    dump_hdr (out->cur_hdr, str, date_fmt);
    if (out->iob->fp) {
	FILE *fp = out->iob->fp;
	fputs (str, fp);
	for (bs=out->cur_hdr->pblockettes; bs!=NULL; bs=bs->next) {
	    BLOCKETTE_HDR *th = NULL;
	    bh = (BLOCKETTE_HDR *)(bs->pb);
	    bl_type = bs->type;
	    bl_len = bs->len;
	    if (bl_type == 2000) {
		/* Ensure we have a copy of the blockette in our wordorder. */
		if (my_wordorder != bs->wordorder) {
		    th = (BLOCKETTE_HDR *)malloc (bs->len);
		    memcpy (th, bh, bl_len);
		    swab_blockette (bs->type, (void *)th, bs->len);
		    bh = th;
		}
		/* Dump the blockette. */
		fprintf (fp, "blockette 2000 (opaque data), len=%d, datalen=%d, "
			 "recno=%d, order=%d (%s), flags=0x%02x, hdr=",
			 ((BLOCKETTE_2000 *)bh)->blockette_len,
			 ((BLOCKETTE_2000 *)bh)->blockette_len - ((BLOCKETTE_2000 *)bh)->data_offset,
			 ((BLOCKETTE_2000 *)bh)->record_num,
			 ((BLOCKETTE_2000 *)bh)->word_order,
			 (((BLOCKETTE_2000 *)bh)->word_order == SEED_BIG_ENDIAN) ? "BIG_ENDIAN" :
			 (((BLOCKETTE_2000 *)bh)->word_order == SEED_LITTLE_ENDIAN) ? "LITTLE_ENDIAN" :
			 "unknown",
			 ((BLOCKETTE_2000 *)bh)->data_flags);
		p = (char *)bh;
		i = (char *)&(((BLOCKETTE_2000 *)bh)->num_hdr_strings) - p;
		p += i + sizeof(((BLOCKETTE_2000 *)bh)->num_hdr_strings);
		for (i=0; i<((BLOCKETTE_2000 *)bh)->num_hdr_strings; i++) {
		    charvncpy(str,p,MAXODHSTRLEN,i);
		    fprintf (fp, "%s%s", ((i>0) ? "," : ""), str);
		}
		fprintf (fp, "\n");
		/* Free temp copy of the blockette if allocated. */
		if (th) free (th);
	    }
	}
	fflush (fp);
    }
}

/************************************************************************/
/*  dump_blk:								*/
/*	Write blockette 2000 summary info info a string, for debugging.	*/
/************************************************************************/
void dump_blk 
   (DATA_HDR	*h,		/* ptr to Data_Hdr structure.		*/
    char	*outstr)	/* write debugging info into string.	*/
{
    BS *bs;
    BLOCKETTE_HDR *bh;
    char *p, str[MAXODHSTRLEN+1];
    int i;
    int bl_type, bl_len;

    if (my_wordorder < 0) get_my_wordorder();
    {
	for (bs=h->pblockettes; bs!=NULL; bs=bs->next) {
	    BLOCKETTE_HDR *th = NULL;
	    bl_type = bs->type;
	    bl_len = bs->len;
	    bh = (BLOCKETTE_HDR *)(bs->pb);
	    if (bl_type == 2000) {
		/* Ensure we have a copy of the blockette in our wordorder. */
		if (my_wordorder != bs->wordorder) {
		    th = (BLOCKETTE_HDR *)malloc (bs->len);
		    memcpy (th, bh, bl_len);
		    swab_blockette (bs->type, (void *)th, bs->len);
		    bh = th;
		}
		/* Dump the blockette. */
		sprintf (outstr, "blockette 2000 (opaque data), len=%d, datalen=%d, "
			 "recno=%d, order=%d (%s), flags=0x%02x, hdr=",
			 ((BLOCKETTE_2000 *)bh)->blockette_len,
			 ((BLOCKETTE_2000 *)bh)->blockette_len - ((BLOCKETTE_2000 *)bh)->data_offset,
			 ((BLOCKETTE_2000 *)bh)->record_num,
			 ((BLOCKETTE_2000 *)bh)->word_order,
			 (((BLOCKETTE_2000 *)bh)->word_order == SEED_BIG_ENDIAN) ? "BIG_ENDIAN" :
			 (((BLOCKETTE_2000 *)bh)->word_order == SEED_LITTLE_ENDIAN) ? "LITTLE_ENDIAN" :
			 "unknown",
			 ((BLOCKETTE_2000 *)bh)->data_flags);
		outstr += strlen(outstr);
		p = (char *)bh;
		i = (char *)&(((BLOCKETTE_2000 *)bh)->num_hdr_strings) - p;
		p += i + sizeof(((BLOCKETTE_2000 *)bh)->num_hdr_strings);
		for (i=0; i<((BLOCKETTE_2000 *)bh)->num_hdr_strings; i++) {
		    charvncpy(str,p,MAXODHSTRLEN,i);
		    sprintf (outstr, "%s%s", ((i>0) ? "," : ""), str);
		    outstr += strlen(outstr);
		}
		sprintf (outstr, "\n");
		outstr += strlen(outstr);
		/* Free temp copy of the blockette if allocated. */
		if (th) free (th);
	    }
	}
    }
}
